from pydantic import BaseModel, Field
from typing import List, Optional, Dict, Literal


class MOSIPRequestedAuth(BaseModel):
    demo: bool = False
    pin: bool = False
    otp: bool = False
    bio: bool = False


class IdentityInfo(BaseModel):
    language: str
    value: str


class DemographicsModel(BaseModel):
    age: str = ""
    dob: str = ""
    name: List[IdentityInfo] = []
    dobType: List[IdentityInfo] = Field(default_factory=list, alias="dob_type")
    gender: List[IdentityInfo] = Field(default_factory=list)
    phoneNumber: str = Field(default="", alias="phone_number")
    emailId: str = Field(default="", alias="email_id")
    addressLine1: List[IdentityInfo] = Field(
        default_factory=list, alias="address_line1"
    )
    addressLine2: List[IdentityInfo] = Field(
        default_factory=list, alias="address_line2"
    )
    addressLine3: List[IdentityInfo] = Field(
        default_factory=list, alias="address_line3"
    )
    location1: List[IdentityInfo] = Field(default_factory=list)
    location2: List[IdentityInfo] = Field(default_factory=list)
    location3: List[IdentityInfo] = Field(default_factory=list)
    postalCode: str = Field(default="", alias="postal_code")
    fullAddress: List[IdentityInfo] = Field(default_factory=list, alias="full_address")
    metadata: Optional[Dict[str, object]] = None


class MOSIPEncryptAuthRequest(BaseModel):
    biometrics: list
    demographics: DemographicsModel
    otp: str
    timestamp: str


class MOSIPBaseRequest(BaseModel):
    id: str
    version: str
    individualId: str
    individualIdType: str
    transactionID: str
    requestTime: str
class MOSIPOtpRequest(MOSIPBaseRequest):
    otpChannel:  List[Literal['phone', 'email']]
    metadata: dict

class MOSIPAuthRequest(BaseModel):
    id: str
    version: str
    individualId: str
    individualIdType: str
    transactionID: str
    requestTime: str
    specVersion: str
    thumbprint: str
    domainUri: str
    env: str
    requestedAuth: MOSIPRequestedAuth = MOSIPRequestedAuth()
    consentObtained: bool
    requestHMAC: str
    requestSessionKey: str
    request: str
    metadata: dict


class BiometricModelDataDigitalIdField(BaseModel):
    serialNo: str
    make: str
    model: str
    type: str
    deviceSubType: str
    deviceProvider: str
    dp: str
    dpId: str
    deviceProviderId: str
    dateTime: str


class BiometricModelDataField(BaseModel):
    digitalId: BiometricModelDataDigitalIdField
    bioType: str
    bioSubType: str
    bioValue: str
    deviceCode: str
    deviceServiceVersion: str
    transactionId: str
    timestamp: str
    purpose: str
    env: str
    version: str
    domainUri: str
    requestedScore: int
    qualityScore: int


class BiometricModel(BaseModel):
    data: BiometricModelDataField
    hash: str
    sessionKey: str
    specVersion: str
    thumbprint: str



from mosip_auth_sdk._authenticator.auth_models import (
    DemographicsModel,
    IdentityInfo,
    BiometricModel,
    BiometricModelDataField,
    BiometricModelDataDigitalIdField,
)

__all__ = [
    "DemographicsModel",
    "IdentityInfo",
    "BiometricModel",
    "BiometricModelDataField",
    "BiometricModelDataDigitalIdField",
]



import requests


class RestUtility:
    def __init__(self, auth_server_url, authorization_header_constant, *, logger):
        self.auth_server_url = auth_server_url
        self.request_headers = {
            "Authorization": authorization_header_constant,
            "content-type": "application/json",
        }
        self.logger = logger

    def get_request(self, path_params=None, headers={}, data=None, cookies=None):
        server_url = self.auth_server_url
        if path_params:
            server_url += path_params
        self.logger.info(
            "Got <GET> Request for URL and Path Params: {}".format(server_url)
        )
        return requests.get(
            server_url,
            headers=headers,
            data=data,
            cookies=cookies,
        )

    def post_request(
        self, path_params=None, additional_headers={}, data=None, cookies=None
    ):
        # is it required to add validation for data is not None?

        server_url = self.auth_server_url
        if path_params:
            if not server_url.endswith("/"):
                server_url += "/"
            server_url += path_params

        if additional_headers:
            self.request_headers.update(additional_headers)

        self.logger.info("Got <POST> Request for URL: {}".format(self.auth_server_url))
        self.logger.debug(f"Final request route = {server_url}")
        self.logger.debug(f"Request Headers = {self.request_headers}")
        return requests.post(
            server_url,
            headers=self.request_headers,
            data=data,
            cookies=cookies,
        )



import string
import secrets
import logging
import sys
import urllib
from datetime import datetime
from typing import Literal, Optional, Dict, TypeAlias, List
from .auth_models import (
    MOSIPAuthRequest,
    DemographicsModel,
    MOSIPEncryptAuthRequest,
    BiometricModel,
    MOSIPBaseRequest,
    MOSIPOtpRequest,
)
from .utils import CryptoUtility, RestUtility
from .exceptions import AuthenticatorException, Errors, AuthenticatorCryptoException

AuthController: TypeAlias = Literal["kyc", "auth"]


class MOSIPAuthenticator:
    """ """
    def __init__(self, *, config, logger=None):
        """ """
        self._validate_config(config)
        if not logger:
            self.logger = self._init_logger(
                file_path=config.logging.log_file_path,
                level=config.logging.loglevel or logging.INFO,
                format=config.logging.log_format,
            )
        else:
            self.logger = logger

        self.auth_rest_util = RestUtility(
            config.mosip_auth_server.ida_auth_url,
            config.mosip_auth.authorization_header_constant,
            logger=self.logger,
        )
        self.crypto_util = CryptoUtility(
            config.crypto_encrypt, config.crypto_signature, self.logger
        )

        self.auth_domain_scheme = config.mosip_auth_server.ida_auth_domain_uri

        self.partner_misp_lk = str(config.mosip_auth.partner_misp_lk)
        self.partner_id = str(config.mosip_auth.partner_id)
        self.partner_apikey = str(config.mosip_auth.partner_apikey)

        self.ida_auth_version = config.mosip_auth.ida_auth_version
        self.ida_auth_request_id_by_controller: Dict[AuthController, str] = {
            "auth": config.mosip_auth.ida_auth_request_demo_id,
            "kyc": config.mosip_auth.ida_auth_request_kyc_id,
            "otp": config.mosip_auth.ida_auth_request_otp_id,
        }
        self.ida_auth_env = config.mosip_auth.ida_auth_env
        self.timestamp_format = config.mosip_auth.timestamp_format
        self.authorization_header_constant = (
            config.mosip_auth.authorization_header_constant
        )

    def genotp(
            self,
            *,
            individual_id,
            individual_id_type,
            txn_id: Optional[str] = '',
            email: Optional[bool] = False,
            phone: Optional[bool] = False,
    ):
        channels = [
            channel
            for (channel, val) in
            (('email', email), ('phone', phone))
            if val
        ]
        if not channels:
            err_msg = Errors.AUT_OTP_001.value
            self.logger.error(err_msg)
            raise AuthenticatorException(Errors.AUT_OTP_001.name, err_msg)

        request = self._get_default_auth_request(
            'otp',
            individual_id=individual_id,
            id_type=individual_id_type,
            txn_id=txn_id,
        )
        request.otpChannel = channels

        path_params = "/".join(
            map(
                urllib.parse.quote,
                (
                    'otp',
                    self.partner_misp_lk,
                    self.partner_id,
                    self.partner_apikey,
                ),
            )
        )
        full_request_json = request.json()
        self.logger.debug(f"{full_request_json=}")
        try:
            signature_header = {
                "Signature": self.crypto_util.sign_auth_request_data(full_request_json)
            }
        except AuthenticatorCryptoException as exp:
            self.logger.error(
                "Failed to Encrypt Auth Data. Error Message: {}".format(exp)
            )
            raise exp

        self.logger.debug(f'Posting to {path_params}')
        response = self.auth_rest_util.post_request(
            path_params=path_params,
            data=full_request_json,
            additional_headers=signature_header,
        )
        self.logger.info("Auth Request for Demographic Completed.")
        return response

    def auth(
        self,
        *,
        individual_id,
        individual_id_type,
        demographic_data: DemographicsModel,
        otp_value: Optional[str] = "",
        biometrics: Optional[List[BiometricModel]] = [],
        consent=False,
    ):
        return self._authenticate(
            controller="auth",
            individual_id=individual_id,
            individual_id_type=individual_id_type,
            demographic_data=demographic_data,
            otp_value=otp_value,
            biometrics=biometrics,
            consent_obtained=consent,
        )

    def kyc(
        self,
        *,
        individual_id,
        individual_id_type,
        demographic_data: DemographicsModel,
        otp_value: Optional[str] = "",
        biometrics: Optional[List[BiometricModel]] = [],
        consent=False,
    ):
        return self._authenticate(
            controller="kyc",
            individual_id=individual_id,
            individual_id_type=individual_id_type,
            demographic_data=demographic_data,
            otp_value=otp_value,
            biometrics=biometrics,
            consent_obtained=consent,
        )

    def decrypt_response(self, response_body):
        r = response_body.get("response")
        session_key_b64 = r.get("sessionKey")
        identity_b64 = r.get("identity")
        # thumbprint should match the SHA-256 hex of the partner certificate
        # thumbprint = response_body.get('thumbprint')
        decrypted = self.crypto_util.decrypt_auth_data(session_key_b64, identity_b64)
        return decrypted

    @staticmethod
    def _validate_config(config):
        if not config.mosip_auth_server.ida_auth_url:
            raise KeyError(
                "Config should have 'ida_auth_url' set under [mosip_auth_server] section"
            )
        if not config.mosip_auth_server.ida_auth_domain_uri:
            raise KeyError(
                "Config should have 'ida_auth_domain_uri' set under [mosip_auth_server] section"
            )

    @staticmethod
    def _init_logger(*, file_path, format, level):
        logger = logging.getLogger(file_path)
        logger.setLevel(level)
        fileHandler = logging.FileHandler(file_path)
        streamHandler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(format)
        streamHandler.setFormatter(formatter)
        fileHandler.setFormatter(formatter)
        logger.addHandler(streamHandler)
        logger.addHandler(fileHandler)
        return logger

    def _get_default_base_request(
            self,
            controller,
            timestamp,
            txn_id,
            individual_id,
            individual_id_type,
    ):
        _timestamp = timestamp or datetime.utcnow()
        timestamp_str = (
            _timestamp.strftime(self.timestamp_format)
            + _timestamp.strftime(".%f")[0:4]
            + "Z"
        )
        transaction_id = txn_id or "".join(
            [secrets.choice(string.digits) for _ in range(10)]
        )
        id = self.ida_auth_request_id_by_controller.get(controller, "")
        if not id:
            err_msg = Errors.AUT_CRY_005.value.format(
                repr(controller),
                " | ".join(self.ida_auth_request_id_by_controller.keys()),
            )
            self.logger.error(f"No id found for {controller}")
            raise AuthenticatorException(Errors.AUT_CRY_005.name, err_msg)

        return MOSIPBaseRequest(
            id=id,
            version=self.ida_auth_version,
            individualId=individual_id,
            individualIdType=individual_id_type,
            transactionID=transaction_id,
            requestTime=timestamp_str,
        )


    def _get_default_auth_request(
        self,
        controller: AuthController,
        *,
        timestamp=None,
        individual_id="",
        txn_id="",
        consent_obtained=False,
        id_type="VID",
    ):
        baserequest = self._get_default_base_request(
            controller, timestamp, txn_id,
            individual_id,
            id_type,
        )
        if controller == 'otp':
            return MOSIPOtpRequest(
                id=baserequest.id,
            	version=baserequest.version,
            	individualId=baserequest.individualId,
            	individualIdType=baserequest.individualIdType,
            	transactionID=baserequest.transactionID,
            	requestTime=baserequest.requestTime,
                otpChannel=[],
                metadata={},
            )
        return MOSIPAuthRequest(
            ## base request
            id=baserequest.id,
            version=baserequest.version,
            individualId=baserequest.individualId,
            individualIdType=baserequest.individualIdType,
            transactionID=baserequest.transactionID,
            requestTime=baserequest.requestTime,
            ## BaseAuthRequestDto
            specVersion=self.ida_auth_version,
            thumbprint=self.crypto_util.enc_cert_thumbprint,
            domainUri=self.auth_domain_scheme,
            env=self.ida_auth_env,
            ## AuthRequestDto
            request="",
            consentObtained=consent_obtained,
            requestHMAC="",
            requestSessionKey="",
            metadata={},
        )

    def _authenticate(
        self,
        *,
        controller: AuthController,
        individual_id: str,
        demographic_data: DemographicsModel,
        otp_value: Optional[str] = "",
        biometrics: Optional[List[BiometricModel]] = [],
        consent_obtained=False,
        individual_id_type=None,
    ):
        """ """
        self.logger.info("Received Auth Request for demographic.")
        auth_request = self._get_default_auth_request(
            controller,
            individual_id=individual_id,
            consent_obtained=consent_obtained,
            id_type=individual_id_type,
        )
        # auth_request.requestedAuth.demo = True
        # auth_request.requestedAuth.otp = bool(otp_value)
        # auth_request.requestedAuth.bio = bool(biometrics)
        request = MOSIPEncryptAuthRequest(
            timestamp=auth_request.requestTime,
            biometrics=biometrics or [],
            demographics=demographic_data,
            otp=otp_value,
        )

        try:
            (
                auth_request.request,
                auth_request.requestSessionKey,
                auth_request.requestHMAC,
            ) = self.crypto_util.encrypt_auth_data(request.json(exclude_unset=True))
        except AuthenticatorCryptoException as exp:
            self.logger.error(
                "Failed to Encrypt Auth Data. Error Message: {}".format(exp)
            )
            raise exp

        path_params = "/".join(
            map(
                urllib.parse.quote,
                (
                    controller,
                    self.partner_misp_lk,
                    self.partner_id,
                    self.partner_apikey,
                ),
            )
        )
        full_request_json = auth_request.json()
        self.logger.debug(f"{full_request_json=}")
        try:
            signature_header = {
                "Signature": self.crypto_util.sign_auth_request_data(full_request_json)
            }
        except AuthenticatorCryptoException as exp:
            self.logger.error(
                "Failed to Encrypt Auth Data. Error Message: {}".format(exp)
            )
            raise exp

        response = self.auth_rest_util.post_request(
            path_params=path_params,
            data=full_request_json,
            additional_headers=signature_header,
        )
        self.logger.info("Auth Request for Demographic Completed.")
        return response



import os
import base64
import json
import logging
import traceback

from cryptography import x509
from cryptography.hazmat.primitives import hashes, asymmetric, ciphers, serialization
from cryptography.hazmat.primitives.serialization import pkcs12, BestAvailableEncryption
from jwcrypto import jwk, jws

from ..exceptions import AuthenticatorCryptoException, Errors


## 4 comes from just how b64 encodes, it considers 3 bytes at one go
## to encode 4 base 64 chars
def _b64pad(s):
    return s + "=" * ((4 - len(s) % 4) % 4)


class CryptoUtility:
    ## TODO: replace encrypt_config in toml with better schema
    def __init__(self, encrypt_config, sign_config, logger, **kwargs):
        if not logger:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger

        self.encrypt_cert_obj = CryptoUtility._get_certificate_obj(
            encrypt_config.encrypt_cert_path, self.logger
        )

        # XXX: make this optional??
        self.decrypt_private_key, *_ = CryptoUtility._get_priv_key_cert(
            encrypt_config.decrypt_p12_file_path,
            str(encrypt_config.decrypt_p12_file_password),
            self.logger,
        )
        self.sign_private_key, self.sign_cert, self.ca_chain = (
            CryptoUtility._get_priv_key_cert(
                sign_config.sign_p12_file_path,
                str(sign_config.sign_p12_file_password),
                self.logger,
            )
        )

        self.sign_priv_key_jws = CryptoUtility._get_jwk_private_key(
            self.sign_private_key, str(sign_config.sign_p12_file_password), self.logger
        )

        self.enc_cert_thumbprint = base64.urlsafe_b64encode(
            self.encrypt_cert_obj.fingerprint(hashes.SHA256())
        )

        # Initalizing OAEP padding algorithm.
        ## https://github.com/mosip/keymanager/blob/17cadbc806b4fc9a80bc147171d6a9be01e9dcc9/kernel/kernel-keymanager-service/src/main/java/io/mosip/kernel/crypto/jce/core/CryptoCore.java#L278
        self.asymmetric_encrypt_padding = asymmetric.padding.OAEP(
            mgf=asymmetric.padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        )
        self.symmetric_key_size = encrypt_config.symmetric_key_size
        self.symmetric_nonce_size = int(encrypt_config.symmetric_nonce_size / 8)
        self.symmetric_gcm_tag_size = int(encrypt_config.symmetric_gcm_tag_size / 8)
        self.algorithm = sign_config.algorithm

    @staticmethod
    def _get_certificate_obj(cert_path, logger):
        # will add not 'None' validation in DynaConf validation API.
        logger.info(
            "Creating certificate Object for the file Path: {}".format(cert_path)
        )
        try:
            with open(cert_path, "rb") as file:
                cert = x509.load_pem_x509_certificate(file.read())
                logger.info("Cretificate Object Creation successful(PEM Format).")
                return cert
        ## XXX: Narrow down the exceptions
        except Exception:
            exp = traceback.format_exc()
            logger.error("Error Reading PEM file format. Error Message: {}".format(exp))

        try:
            with open(cert_path, "rb") as file:
                cert = x509.load_der_x509_certificate(file.read())
                logger.info("Cretificate Object Creation successful(DER Format).")
                return cert
        ## XXX: Narrow down the exceptions
        except Exception:
            exp = traceback.format_exc()
            logger.error(
                "Error Reading both PEM/DER file format. Error: {}".format(exp)
            )
            raise AuthenticatorCryptoException(
                Errors.AUT_CRY_001.name, Errors.AUT_CRY_001.value.format(cert_path)
            )

    @staticmethod
    def _get_priv_key_cert(p12_file_path, p12_file_pass, logger):
        logger.info("Reading P12 file. File Path: {}".format(p12_file_path))
        try:
            with open(p12_file_path, "rb") as file:
                pem_bytes = file.read()
                return pkcs12.load_key_and_certificates(
                    pem_bytes, bytes(p12_file_pass, "utf-8")
                )
        ## XXX: Narrow down the exceptions
        except Exception:
            exp = traceback.format_exc()
            logger.error(
                "Error Loading P12 file to create objects. Error: {}".format(exp)
            )
            raise AuthenticatorCryptoException(
                Errors.AUT_CRY_002.name, Errors.AUT_CRY_002.value.format(p12_file_path)
            )

    @staticmethod
    def _get_jwk_private_key(priv_key_obj, key_password, logger):
        key_pwd_bytes = bytes(key_password, "utf-8")
        priv_key_pem = priv_key_obj.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=BestAvailableEncryption(key_pwd_bytes),
        )
        logger.info("Creating JWK key for JWS signing.")
        return jwk.JWK.from_pem(priv_key_pem, password=key_pwd_bytes)

    def _asymmetric_encrypt(self, aes_random_key):
        self.logger.debug("Encrypting the AES Random Key.")
        public_key_obj = self.encrypt_cert_obj.public_key()
        return public_key_obj.encrypt(aes_random_key, self.asymmetric_encrypt_padding)

    def _asymmetric_decrypt(self, encrypted_data):
        self.logger.debug("Asymmetric Decryption")

        ## algo = RSA/ECB/OAEPWITHSHA-256ANDMGF1PADDING
        ##
        symmetric_key = self.decrypt_private_key.decrypt(
            encrypted_data, self.asymmetric_encrypt_padding
        )
        return symmetric_key

    def _symmetric_encrypt(self, data, key):
        self.logger.debug("Encrypting the Auth Data using AES Key.")
        iv = os.urandom(self.symmetric_nonce_size)
        aes_encryptor_obj = ciphers.Cipher(
            ciphers.algorithms.AES(key),
            ciphers.modes.GCM(iv, tag=None, min_tag_length=self.symmetric_gcm_tag_size),
        ).encryptor()

        enc_data = aes_encryptor_obj.update(data) + aes_encryptor_obj.finalize()
        enc_data_final = enc_data + aes_encryptor_obj.tag + iv
        return enc_data_final

    def _symmetric_decrypt(self, data, key):
        self.logger.debug("Decrypting the Auth Data using AES Key.")
        len_iv = self.symmetric_nonce_size
        len_tag = self.symmetric_gcm_tag_size

        iv = data[-len_iv:]
        tag = data[-(len_tag + len_iv) : -len_iv]
        enc_data = data[: -(len_tag + len_iv)]
        aes_decryptor_obj = ciphers.Cipher(
            ciphers.algorithms.AES(key),
            ciphers.modes.GCM(iv, tag, min_tag_length=self.symmetric_gcm_tag_size),
        ).decryptor()
        data = aes_decryptor_obj.update(enc_data) + aes_decryptor_obj.finalize()
        return data

    def decrypt_auth_data(self, session_key_b64, encrypted_identity_b64):
        ## sometimes response doesn't pad the data correctly
        ## temporary workaround
        session_key_b64_padded = _b64pad(session_key_b64)
        encrypted_identity_b64_padded = _b64pad(encrypted_identity_b64)
        session_key = base64.urlsafe_b64decode(session_key_b64_padded)
        encrypted_identity = base64.urlsafe_b64decode(encrypted_identity_b64_padded)
        sym_key = self._asymmetric_decrypt(session_key)
        identity = self._symmetric_decrypt(encrypted_identity, sym_key)
        ## TODO: decide whether  we should be doing this here
        return json.loads(identity)

    def encrypt_auth_data(self, auth_data):
        self.logger.info("Request for Auth Data Encryption.")
        if isinstance(auth_data, str):
            auth_data_bytes = auth_data.encode("UTF-8")
        elif isinstance(auth_data, bytes):
            auth_data_bytes = auth_data
        else:
            raise ValueError("Unrecognised type")

        try:
            # Generate a random AES Key and encrypt Auth Request Data using the generated random key.
            aes_key = os.urandom(int(self.symmetric_key_size / 8))
            encrypted_auth_data = self._symmetric_encrypt(auth_data_bytes, aes_key)
            encrypted_auth_b64_data = base64.urlsafe_b64encode(
                encrypted_auth_data
            ).decode("UTF-8")
            self.logger.info("Generating AES Key and encrypting Auth Data Completed.")

            # Encrypt the random generated key using the IDA partner certificate received from IDA server.
            encrypted_aes_key = self._asymmetric_encrypt(aes_key)
            encrypted_aes_key_b64 = base64.urlsafe_b64encode(encrypted_aes_key).decode(
                "UTF-8"
            )
            self.logger.info("Encrypting Random AES Key Completed.")

            # Generate SHA256 for the Auth Request Data
            sha256_hash_obj = hashes.Hash(hashes.SHA256())
            sha256_hash_obj.update(auth_data_bytes)
            auth_data_hash = sha256_hash_obj.finalize().hex().upper().encode("UTF-8")
            enc_auth_data_hash = self._symmetric_encrypt(auth_data_hash, aes_key)
            enc_auth_data_hash_b64 = base64.urlsafe_b64encode(
                enc_auth_data_hash
            ).decode("UTF-8")
            self.logger.info("Generation of SHA256 Hash for the Auth Data completed.")

            return (
                encrypted_auth_b64_data,
                encrypted_aes_key_b64,
                enc_auth_data_hash_b64,
            )
        ## XXX: Narrow down the exceptions
        except Exception:
            exp = traceback.format_exc()
            self.logger.error(
                "Error encrypting Auth Data. Error Message: {}".format(exp)
            )
            raise AuthenticatorCryptoException(
                Errors.AUT_CRY_003.name, Errors.AUT_CRY_003.value
            )

    def sign_auth_request_data(self, auth_request_data):
        self.logger.info("Request for Sign Auth Request Data.")
        try:
            jws_object = jws.JWS(auth_request_data.encode("UTF-8"))
            jws_object.add_signature(
                self.sign_priv_key_jws,
                None,
                json.dumps(
                    {
                        "alg": self.algorithm,
                        "x5c": [
                            base64.encodebytes(
                                self.sign_cert.public_bytes(
                                    encoding=serialization.Encoding.PEM
                                )
                            ).decode("UTF-8")
                        ],
                    }
                ),  # Protected header attributes, is IDA checking for protected attributes?
                json.dumps(
                    {
                        "kid": base64.encodebytes(
                            self.sign_cert.fingerprint(hashes.SHA256())
                        ).decode("UTF-8")
                    }
                ),  # UnProtected Header attributes.
            )

            jws_signature = jws_object.serialize(compact=True).split(".")
            self.logger.info("Generation for JWS Signature completed.")
            return jws_signature[0] + ".." + jws_signature[2]
        except Exception:
            exp = traceback.format_exc()
            self.logger.error("Error Signing Auth Data. Error Message: {}".format(exp))
            raise AuthenticatorCryptoException(
                Errors.AUT_CRY_004.name, Errors.AUT_CRY_004.value
            )



from enum import Enum


class AuthenticatorException(Exception):
    def __init__(self, error_code, error_message):
        self.error_code = error_code
        self.error_message = error_message
        super().__init__(self.error_message)


class AuthenticatorCryptoException(Exception):
    def __init__(self, error_code, error_message):
        self.error_code = error_code
        self.error_message = error_message
        super().__init__(self.error_message)


class Errors(Enum):
    AUT_CRY_001 = (
        "Error Parsing Encryption Certificate provided in config file. File Name: {}"
    )
    AUT_CRY_002 = "Error Reading P12 file provided in config file. File Name: {}"
    AUT_CRY_003 = "Error Encrypting Auth Data."
    AUT_CRY_004 = "Error Signing Auth Request Data."
    AUT_CRY_005 = "Controller Method Not Found For Method {}. Supported Methods are {}"

    AUT_BAS_001 = "Not Able to process auth request."

    AUT_OTP_001 = "No channels found. Please Pass otp=True or phone=True"
